/*----------------------------------------------------------------------------
 *
 *   Copyright (C) 2021 Antonio Augusto Alves Junior
 *
 *   This file is part of RandomIterator.
 *
 *   RandomIterator is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   RandomIterator is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with RandomIterator.  If not, see <http://www.gnu.org/licenses/>.
 *
 *---------------------------------------------------------------------------*/
/*
 * benchmark_engine.hpp
 *
 *  Created on: 25/02/2021
 *      Author: Antonio Augusto Alves Junior
 */


#define CATCH_CONFIG_MAIN  // This tells Catch to provide a main() - only do this in one cpp file
#define CATCH_CONFIG_ENABLE_BENCHMARKING
#include <catch.hpp>

#include <cstdio>
#include <cstdint>

//hydra
#include <random_iterator/Stream.hpp>
#include <random>

//set a global seed
static const uint64_t default_seed= 0x548c9decbce65295  ;

TEST_CASE("Performance of counter based PRNGs and corresponding iterators") {


	random_iterator::squares3_64        RNG_squares3(default_seed);
	random_iterator::squares4_64        RNG_squares4(default_seed);
	random_iterator::squares3_128       RNG_squares3_long(default_seed);
	random_iterator::squares4_128       RNG_squares4_long(default_seed);
	random_iterator::threefry           RNG_threefry(default_seed);
	random_iterator::philox             RNG_philox(default_seed);
	random_iterator::ars                RNG_ars(default_seed);

	std::mt19937         RNG_mt19937(default_seed );
	std::ranlux48        RNG_ranlux48(default_seed);

	SECTION( "Benchmarks: RNG::operator()" ){


	BENCHMARK("random_iterator::squares3") {
		return RNG_squares3();
	};

	BENCHMARK("random_iterator::squares3_long") {
			return RNG_squares3_long();
	};

	BENCHMARK("random_iterator::squares4") {
		return RNG_squares4();
	};


	BENCHMARK("random_iterator::squares4_long") {
		return RNG_squares4_long();
	};


	BENCHMARK("random_iterator::threefry") {
		return RNG_threefry();
	};

	BENCHMARK("random_iterator::philox") {
		return RNG_philox();
	};

	BENCHMARK("random_iterator::ars") {
		return RNG_ars();
	};

	BENCHMARK("std::mt19937") {
		return RNG_mt19937();
	};

	BENCHMARK("std::ranlux48") {
		return RNG_ranlux48();
	};
	}


	SECTION( "Benchmarks: RNG::discard(1)" )
	{

		static uint32_t jump_size=1;

	BENCHMARK("random_iterator::squares3") {
		RNG_squares3.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4") {
		return RNG_squares4.discard(jump_size);
	};


	BENCHMARK("random_iterator::squares3_long") {
		RNG_squares3_long.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4_long") {
		return RNG_squares4_long.discard(jump_size);
	};



	BENCHMARK("random_iterator::threefry") {
		return RNG_threefry.discard(jump_size);
	};

	BENCHMARK("random_iterator::philox") {
		return RNG_philox.discard(jump_size);
	};

	BENCHMARK("random_iterator::ars") {
		return RNG_ars.discard(jump_size);
	};

	BENCHMARK("std::mt19937") {
		return RNG_mt19937.discard(jump_size);
	};

	BENCHMARK("std::ranlux48") {
		return RNG_ranlux48.discard(jump_size);
	};
	}

	SECTION( "Benchmarks: RNG::discard(10)" )
	{
	static uint32_t jump_size=10;

	BENCHMARK("random_iterator::squares3") {
		return RNG_squares3.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4") {
		return RNG_squares4.discard(jump_size);
	};


	BENCHMARK("random_iterator::squares3_long") {
		RNG_squares3_long.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4_long") {
		return RNG_squares4_long.discard(jump_size);
	};

	BENCHMARK("random_iterator::threefry") {
		return RNG_threefry.discard(jump_size);
	};

	BENCHMARK("random_iterator::philox") {
		return RNG_philox.discard(jump_size);
	};

	BENCHMARK("random_iterator::ars") {
		return RNG_ars.discard(jump_size);
	};

	BENCHMARK("std::mt19937") {
		return RNG_mt19937.discard(jump_size);
	};

	BENCHMARK("std::ranlux48") {
		return RNG_ranlux48.discard(jump_size);
	};
	}


	SECTION( "Benchmarks: RNG::discard(100)" )
	{
	static uint32_t jump_size=100;

	BENCHMARK("random_iterator::squares3") {
		return RNG_squares3.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4") {
		return RNG_squares4.discard(jump_size);
	};


	BENCHMARK("random_iterator::squares3_long") {
		RNG_squares3_long.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4_long") {
		return RNG_squares4_long.discard(jump_size);
	};

	BENCHMARK("random_iterator::threefry") {
		return RNG_threefry.discard(jump_size);
	};

	BENCHMARK("random_iterator::philox") {
		return RNG_philox.discard(jump_size);
	};

	BENCHMARK("random_iterator::ars") {
		return RNG_ars.discard(jump_size);
	};

	BENCHMARK("std::mt19937") {
		return RNG_mt19937.discard(jump_size);
	};

	BENCHMARK("std::ranlux48") {
		return RNG_ranlux48.discard(jump_size);
	};
	}

	SECTION( "Benchmarks: RNG::discard(1000)" )
	{
	static uint32_t jump_size=1000;

	BENCHMARK("random_iterator::squares3") {
		return RNG_squares3.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4") {
		return RNG_squares4.discard(jump_size);
	};


	BENCHMARK("random_iterator::squares3_long") {
		RNG_squares3_long.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4_long") {
		return RNG_squares4_long.discard(jump_size);
	};

	BENCHMARK("random_iterator::threefry") {
		return RNG_threefry.discard(jump_size);
	};

	BENCHMARK("random_iterator::philox") {
		return RNG_philox.discard(jump_size);
	};

	BENCHMARK("random_iterator::ars") {
		return RNG_ars.discard(jump_size);
	};

	BENCHMARK("std::mt19937") {
		return RNG_mt19937.discard(jump_size);
	};

	BENCHMARK("std::ranlux48") {
		return RNG_ranlux48.discard(jump_size);
	};
	}


	SECTION( "Benchmarks: RNG::discard(10000)" )
	{
	static uint32_t jump_size=10000;

	BENCHMARK("random_iterator::squares3") {
		return RNG_squares3.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4") {
		return RNG_squares4.discard(jump_size);
	};


	BENCHMARK("random_iterator::squares3_long") {
		RNG_squares3_long.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4_long") {
		return RNG_squares4_long.discard(jump_size);
	};

	BENCHMARK("random_iterator::threefry") {
		return RNG_threefry.discard(jump_size);
	};

	BENCHMARK("random_iterator::philox") {
		return RNG_philox.discard(jump_size);
	};

	BENCHMARK("random_iterator::ars") {
		return RNG_ars.discard(jump_size);
	};

	BENCHMARK("std::mt19937") {
		return RNG_mt19937.discard(jump_size);
	};

	BENCHMARK("std::ranlux48") {
		return RNG_ranlux48.discard(jump_size);
	};

	}


	SECTION( "Benchmarks: RNG::discard(100000)" )
	{
	static uint32_t jump_size=100000;

	BENCHMARK("random_iterator::squares3") {
		return RNG_squares3.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4") {
		return RNG_squares4.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares3_long") {
		RNG_squares3_long.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4_long") {
		return RNG_squares4_long.discard(jump_size);
	};

	BENCHMARK("random_iterator::threefry") {
		return RNG_threefry.discard(jump_size);
	};

	BENCHMARK("random_iterator::philox") {
		return RNG_philox.discard(jump_size);
	};

	BENCHMARK("random_iterator::ars") {
		return RNG_ars.discard(jump_size);
	};

	BENCHMARK("std::mt19937") {
		return RNG_mt19937.discard(jump_size);
	};

	BENCHMARK("std::ranlux48") {
		return RNG_ranlux48.discard(jump_size);
	};
	}


	SECTION( "Benchmarks: RNG::discard(1000000)" )
	{
	static uint32_t jump_size=1000000;

	BENCHMARK("random_iterator::squares3") {
		return RNG_squares3.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4") {
		return RNG_squares4.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares3_long") {
		RNG_squares3_long.discard(jump_size);
	};

	BENCHMARK("random_iterator::squares4_long") {
		return RNG_squares4_long.discard(jump_size);
	};

	BENCHMARK("random_iterator::threefry") {
		return RNG_threefry.discard(jump_size);
	};

	BENCHMARK("random_iterator::philox") {
		return RNG_philox.discard(jump_size);
	};

	BENCHMARK("random_iterator::ars") {
		return RNG_ars.discard(jump_size);
	};

	BENCHMARK("std::mt19937") {
		return RNG_mt19937.discard(jump_size);
	};

	BENCHMARK("std::ranlux48") {
		return RNG_ranlux48.discard(jump_size);
	};
	}
}

